; C64-Pong 1.0-beta

; Das folgende Pong-Spiel wurde im Rahmen eines Projektseminars im Sommersemseter 2018 an der Medienwissenschaft
; der Humboldt-Universität zu Berlin entwickelt. Ziel des Projektes war es, ein möglichst einfaches Pong zu programmieren.
; Daher wurde hier nicht bare metal Assembler programmiert, sondern möglichst auf Routinen des BASIC-ROMs zuirückgegriffen.
; Ebenso wurde auf Pixelgrafik verzichtet und anstelle dessen PETSCII-Zeichen für Schläger und Ball und die Bildschirm-
; auflösung des Textbildschirms (40*25) verwendet.
; Im Zentrum der Fragestellung stand das "reverse engieering" der Spiel-I/O und vor allem der Ball-Physik. Ergänzend wurde 
; eine Soundausgabe auf Basis der originalen Pong-Sounds hinzugefügt.


; BASIC-Autostartroutine
*= $801
.word (+), 10
.null $9e, "49152"
+ .word 0

*=$C000								; Startadresse

; INITIALISIERUNG
                                                                           
; Systemvariablen:
	SID=$d400
	RND=$00a2						; Element des BASIC-Timers für Zufallszahlen
	BSOUT=$ffd2						; Gibt Zeichen aus a an Cursorposition aus
	PLOT=$fff0						; Positioniert Textcurser an x,y
	GETIN=$ffe4						; liest Zeichen von Tastatur in a ein
	JOY1=$dc01						; CIA-Port Joystick 1
	JOY2=$dc00						; CIA-Port Joystick 2
; Spielvariablen definieren:
	schlaeger1_xpos=$CF00			; oberer Schläger
	schlaeger2_xpos=$CF01			; unterer Schläger
	ball_x=$CF02					; X-Bildschirmkoordinate des Balls
	ball_y=$CF03					; Y-Bildschirmkoordinate des Balls
	ball_dir=$CF04					; vertikale Ballrichtung (0=hoch; 1=runter)
	ball_speed=$CF05				; vertikale Ballbeschleunigung
	ball_angle=$CF06				; horizontale Ballpositionsänderung (2:x-2 - 1:x-1 - 3:x+0 - 4:x+1 - 5:x+2)
	punkte_player1=$CF07			; Punkte oberer Spieler
	punkte_player2=$CF08			; Punkte unterer Spieler
	restart_flag=$CF09				; Flag für Satz-Gewinn
	win_flag=$CF0A					; Flag für Spielgewinn
	wait_ball_flag=$CF0B			; Flag für Ballgeschiwndigkeit
	counter_ball=$CF0C				; Zähler für Ball-Wartezeit
	return=$CF0D					; Speicher für Return-Position des Balls
	ball_delta=$CF0E				; Delta-x für Ballabweichung
	soundloop=$CF0F					; Länge der Sounds in 1/60 Sekunden: wall=1, paddle=1, point=14
	soundtune=$CF10					; wall-sound=0, paddle-sound=2, score-sound=4 (Zahlen sind Offsets für sound_table)
	wait_bat_flag=$CF11				; Flag für Schlägergeschiwndigkeit
	counter_bat=$CF12				; Zähler für Schläger-Wartezeit
; Spielvariablen initialisieren:
	endscore=#57					; Gewinnpunktzahl (49-57 "0"-"9")
	lda #48							; Punktestände auf "0" zurücksetzen ...
	sta punkte_player1				; ... bei Spieler 1
	sta punkte_player2				; ... bei Spieler 2
restart:							; Einsprungstelle nach Ballverlust
	lda #0							; Zurücksetzen von ...
	sta restart_flag				; ... Restart-Flag
	sta win_flag					; ... Win-Flag
	sta wait_ball_flag				; ... Wait-Flag Ball
	sta wait_bat_flag				; ... Wait-Flag Schläger
	sta soundloop					; ... Soundausgabe
	lda #17							; Schläger 1 ...
	sta schlaeger1_xpos				; ... an Position 5
	lda #17							; Schläger 2 ...
	sta schlaeger2_xpos				; ... an Position 30
; Ball vorbereiten:	
	lda #20							; Ball ...
	sta ball_x						; ... auf dem Bildschirm ....
	lda #12							; ... positionieren:
	sta ball_y						; Zeile 20, Spalte 12
	lda RND							; Startrichtung:
	and #1							; oberste 7 Bit maskieren (Paritätsbit von RND für 0 oder 1) ...
	sta ball_dir					; ... speicherm
	lda #20							; Einwurf-Ballgeschwindigkeit ...
	sta ball_speed					; ... für Watchdog 
	lda #3							; Ballwinkel:
	sta ball_angle					; 3 (180° [gerade])
; Spielfeld vorbereiten:
	lda #$0							; grauer ...
	sta $d020						; ... Bildrahmen ...
	lda #$b							; ... schwarzer ...
	sta $d021						; ... Bildhintergrund
	lda #1							; weiße ...
	sta $0286						; Spielelemente
	lda #147						; Bildschirm ...
	jsr BSOUT						; ... löschen

; HAUPTSCHLEIFE
main:
	jsr play_sound					; Müssen Sounds ausgegeben werden?		
	jsr count_ball					; Kick Watchdog für Ballflug-Verzögerung
	jsr count_bat					; Kick Watchdog für Schläger-Verzögerung
	jsr scores_out					; Scores anzeigen
	jsr schlaeger					; Schläger zeichnen
	jsr joyin						; Joysticks prüfen
	jsr ball_erase					; Alten Ball vom Bildschirm löschen	
	jsr win							; Spiel gewonnen?
		lda win_flag				; Falls ja, ...
		cmp #1						; ... Spiel ...
		beq the_end					; zuende.
	jsr ball_move					; Ballposition neu berechnen
		lda restart_flag			; Ball ist jetzt im Aus?
		cmp #1						; Falls ja, ...
		beq restart					; ... restart
	jsr ball_bat					; Kollisionsprüfung: Schläger
	jsr ball_draw					; Jetzt den Ball an neue Position zeichnen		
jmp main
the_end: jmp $ff99					; Programmende: Zurück ins BASIC	

; BALL-ROUTINEN

; Ballbewegung berechnen
ball_move:							; Neue horizontale und vertikale Ballposition auf Basis der Trefferzone setzen			
	lda wait_ball_flag				; Wait-Flag gesetzt?
	beq fertig_move					; Falls nein: keine neue Ballposition berechnen
	ball_vertical:					; neue vertikale Ballbewegung (ball_y)	
		lda ball_dir				; aktuelle vertikale Flugrichtung
		beq ball_up					; Flug nach oben (=0)?
		bne ball_down				; Flug nach unten (<>0)?
		ball_up:
			dec ball_y				; Subtrahiere 1 von der y-Position des Balls ...
			beq ball_out_player1	; ... und prüfe, ob Ball jetzt auf der Auslinie (y=0) ...
			jmp ball_horizontal		; ... sonst weiter zur horizontalen Berechnung
		ball_down:			
			inc ball_y				; Addiere 1 zur y-Position des Balls und ...
			lda ball_y				; ... prüfe y-Position ...
			cmp #24					; ... ob Ball ...
			beq ball_out_player2	; ... auf der Auslinie (y=24)
	ball_horizontal:				; Neue horizontale Ballbewegung (ball_x) setzen
		ldy ball_angle				; aktuellen Winkel laden ...
		dey							; (damit in delta_table ab Eintrag y=0 geladen wird)
		lda delta_table,y			; ... und einem Summanden/Minuenden aus delta_table ...
		sta ball_delta				; ... zuordnen
		lda ball_angle				; Aktuellen Winkel noch einmal laden
		cmp #1						; Für die linken ...
		beq ball_left				; ... zwei Schlägersegmente ...
		cmp #2						; ... Winkel subrahieren
		beq ball_left
		cmp #3						; Für das mittlere Schlägersegment ... 
		beq ball_straigt			; ... keine Änderung
		cmp #4						; Für die rechten ...
		beq ball_right				; ... zwei Schlägersegmente ...
		cmp #5						; ... Winkel addieren
		beq ball_right
		ball_left:					; Flug nach links:
			lda ball_x				; Von X-Position des Balls ...
			sec
			sbc ball_delta			; ... Delta-X subtrahieren
			cmp #0					; Ball schon ...
			beq ball_wall_left		; ... am linken Bildrand (x=0)?
			bmi ball_wall_left		; ... hinter dem linken Bildrand (x<0)?
			sta ball_x				; Neue X-Position sichern ...
		jmp fertig_move				; ... und ball_move beenden.
		ball_straigt:				; Gerader Flug:
			jmp fertig_move			; Ohne Änderung ball_move beenden.
		ball_right:					; Flug nach rechts:
			lda ball_x				; Delta-x entspricht dem Winkel-Wert ...
			adc ball_delta			; ... und kann deshalb direkt subtrahiert werden
			sta ball_x				; Neue X-Position sichern.
			cmp #39					; Prüfen, ob ...
			bcs ball_wall_right		; ... Ball am rechten Bildrand (x=39)
			bpl ball_wall_right		; ... Ball hinter dem rechten Bildrand (x>39)
	fertig_move:		
rts									; zurück nach main

; Ball im Aus & Punktgewinn
ball_out_player1: 					; Punktgewinn Spieler 1
		inc punkte_player1			; Punkt für Player 1
		lda #1						; Restart-Flag
		sta restart_flag			; setzen
		jmp out_effekt				; "Ball im Aus"-Animation
ball_out_player2:					; Punktgewinn Spieler 2
		inc punkte_player2			; Punkt für Player 2
		lda #1						; Restart-Flag
		sta restart_flag			; setzen
	out_effekt:						; "Ball im Aus"-Animation:
		lda #4						; Sound für Score
		sta soundtune
		lda #14
		sta soundloop		
		jsr wait					; Verzögerungszeit
jmp fertig_move						; Zurück und zurück nach main

; Ballkollision mit der Wand
ball_wall_left:						; Kollision mit linker Wand:
	lda #0							; x-Position des Balls ...
	sta ball_x						; ... auf linken Rand setzen (zur Sicherheit, falls x<0 war)
	jmp ball_bounce					; Neue Flugrichtung
ball_wall_right:					; Kollision mit rechter Wand
	lda #39							; x-Position des Balls ...
	sta ball_x						; ... auf rechten Rand setzen (zur Sicherheit, falls x>39 war)
ball_bounce:						; Neue Flugrichtung:
	ldy ball_angle					; Alter ball_angle als offset ...
	dey								; (damit in ball_table ab Eintrag y=0 geladen wird)
	lda ball_table,y				; ... zum Lesen der neuen ball_angles aus Tabelle
	sta ball_angle					; Neuen Winkel sichern
	lda #0							; Sound für Wall-Ball
	sta soundtune					
	lda #2
	sta soundloop
jmp fertig_move						; Zurück und zurück nach main


; Ballkollision mit dem Schläger prüfen
ball_bat:
	lda ball_y						; vertikale Ballposition holen
	cmp #1							; Prüfung, ob ...
	beq treffer_player1				; ...  Ball auf oberer Schlägerzeile
	cmp #23							; Prüfung, ob ...
	beq treffer_player2				; ...  Ball auf unterer Schlägerzeile
	rts								; ... sonst zurück nach main
	treffer_player1:				; Kollisionspüfung mit Schläger 1
		lda #2						; (Ball später wieder vor ...
		sta return					; ... dem Schläger positionieren)
		lda schlaeger1_xpos			; 1. (linkes) Element von Schläger 1 laden ...
		jmp treffer_zone			; ... und weiter zur Trefferzonenprüfung
	treffer_player2:				; Kollisionsprüfung mit Schläger 2
		lda #22						; (Ball später wieder vor ...
		sta return					; ... dem Schläger positionieren)
		lda schlaeger2_xpos			; 1. (linkes) Element von Schläger 2 laden ...
	treffer_zone:					; Ermitteln der Trefferzone:
		adc #5						; ganz rechtes Schlägerelement
		sec
		sbc ball_x					; Ball- und Schlägerposition zum Vergleichen normalisieren
		cmp #5						; Kollision mit ganz linkem Element?
		beq getroffen
		cmp #4						; Kollision mit zweitem Element von links?
		beq getroffen
		cmp #3						; Kollision mit mittlerem Element?
		beq getroffen
		cmp #2						; Kollision mit zweitem Element von rechtes?
		beq getroffen
		cmp #1						; Kollision mit ganz rechtem Element?
		beq getroffen
		rts							; Keine Kollision mit einem der Schläger: zurück nach main
	getroffen:
		tay							; Trefferzone als Offest laden
		dey							; (damit in treffer_table ab Eintrag y=0 geladen wird)
		lda treffer_table,y			; Wert aus Tabelle holen ...
		sta ball_angle				; ... und als neuen Winkel speichern
		lda speed_table,y			; vertikale Beschleunigung ...
		sta ball_speed				; ... anpassen.
		lda return					; Ball auf ...
		sta ball_y					; ... vorbereitete Return-Position setzen
		lda ball_dir				; Ballrichtung ...
		eor #1						; ... umkehren ...
		sta ball_dir				; ... und speichern
		lda #2						; Sound für Schläger getroffen
		sta soundtune					
		lda #2
		sta soundloop
rts
 
; ROUTINEN ZUR GEWINN-ERMITTLUNG

; Spielgewinn
win:								; Prüft, ob ein Spieler gewonnen hat	
	lda punkte_player1				; Spieler 1 ...
	cmp endscore					; ... schon Endscore erreicht?
	beq player1_win					; ja!
	lda punkte_player2				; Spieler 2 ...
	cmp endscore					; ... schon Endscore erreicht?
	beq player2_win					; ja!
	rts								; ... sonst zurück nach main
	player1_win:
									;;; Hier kann noch eine Animation o.ä. rein
	lda #1							; Setzen des ...
	sta win_flag					; ... Gewinn-Flags
	rts
	player2_win:
									;;; Hier kann noch eine Animation o.ä. rein
	lda #1							; Setzen des ...
	sta win_flag					; ... Gewinn-Flags
	rts

; EINGABE-ROUTINEN

joyin:								; Spielereingaben prüfen:
;	lda wait_bat_flag				; Schläger schon bewegbar?
;	beq joy_end						; Nein? Dann zurück
; 	jsr joywait						; EXPERIMENTELL: Warteschleife
	lda JOY1						; Joystick 1 ...
	cmp #251						; ... nach links?
	beq	sp1links				
	cmp #247						; ... nach rechts?
	beq sp1rechts				
	lda JOY2						; Joystick 2 ...
	cmp #123						; ... nach links?
	beq sp2links				
	cmp #119						; ... nach rechts?
	beq sp2rechts				 
	jmp joy_end						; Keine (erkennbare) Bewergung: Prüfung beenden
	sp1links:
		lda schlaeger1_xpos			; Schläger 1 schon ganz links? ...
		bmi joy_end					; ... dann stopp!
		dec schlaeger1_xpos			; else: Position Schläger 1 nach links
	jmp joy_end
	sp1rechts:
		lda #34						; Schläger 1 schon ...
		cmp schlaeger1_xpos			; ... ganz rechts? ...
		beq joy_end					; ... dann stopp!
		inc schlaeger1_xpos			; else: Position Schläger 1 nach rechts
	jmp joy_end
	sp2links:
		lda schlaeger2_xpos			; Schläger 2 schön ganz links? ...
		bmi joy_end					; ... dann stopp!
		dec schlaeger2_xpos			; else: Position Schläger 2 nach links
	jmp joy_end
	sp2rechts:
		lda #34						; Schläger 1 schon ...
		cmp schlaeger2_xpos			; ... ganz rechts? ...
		beq joy_end					; ... dann stopp!	
		inc schlaeger2_xpos			; else: Position Schläger 2 nach rechts
	joy_end:						; Einer der Schläger berührt eine Wand
rts		

; Verzögerungsschleifen

count_ball:							; Watch-Dog für Ballflug-Verzögerung
	lda #0							; Lösche ...
	sta wait_ball_flag					; ... Wait-Flag
	inc counter_ball				; Erhöhe Speed-Counter ...
	lda counter_ball				; Counter ...
	cmp ball_speed					; ... schon voll?
	beq set_wait_ball_flag			; falls ja: Wait-Flag setzen
	rts								; falls nein: zurück
	set_wait_ball_flag:
	lda #0							; Lösche ...
	sta counter_ball				; ... Speed-Flag und ...
	lda #1							; ... setze ...
	sta wait_ball_flag				; ... Wait-Flag
rts

count_bat:							; Watch-Dog für Schläger-Verzögerung
	lda #0							; Lösche ...
	sta wait_bat_flag				; ... Wait-Flag
	inc counter_bat					; Erhöhe Speed-Counter ...
	lda counter_bat					; Counter ...
	cmp #5							; ... schon voll?
	beq set_wait_bat_flag			; falls ja: Wait-Flag setzen
	rts								; falls nein: zurück
	set_wait_bat_flag:
	lda #0							; Lösche ...
	sta counter_bat					; ... Speed-Flag und ...
	lda #1							; ... setze ...
	sta wait_bat_flag				; ... Wait-Flag
rts

wait:								; Warteschleife mit Flackereffekt
	ldx #$ff
	wait_loop:
		jsr play_sound				; Sound weiterspielen?
		stx $d020
		dex
	bne wait_loop
rts

; AUSGABE-ROUTINEN

; Ausgabe der Schläger
schlaeger:
									; Oberer Schläger:
	clc								; PLOT für Ausgabe vorbereiten
	ldy schlaeger1_xpos				; x-Position laden
	ldx #1							; oberste Zeile
	jsr PLOT						; Positionieren
	lda #32							; Zeichen links neben ...
	jsr BSOUT						; ... dem Schläger löschen.
	lda #102						; Schläger Spieler 2 ...
	jsr BSOUT						; Schläger Element 1
	jsr BSOUT						; Schläger Element 2
	jsr BSOUT						; Schläger Element 3
	jsr BSOUT						; Schläger Element 4
	jsr BSOUT						; Schläger Element 5
	lda #32							; Zeichen rechts neben ...
	jsr BSOUT						; ... dem Schläger löschen.							
									; Unterer Schläger:
	clc								; PLOT für Ausgabe vorbereiten		
	ldy schlaeger2_xpos				; x-Position laden
	ldx #23							; vorletzte Zeile
	jsr PLOT						; Positionieren
	lda #32							; Zeichen links neben ...
	jsr BSOUT						; ... dem Schläger löschen.
	lda #101						; Schläger Spieler 2 ...
	jsr BSOUT						; Schläger Element 1
	jsr BSOUT						; Schläger Element 2
	jsr BSOUT						; Schläger Element 3
	jsr BSOUT						; Schläger Element 4
	jsr BSOUT						; Schläger Element 5
	lda #32							; Zeichen rechts neben ...
	jsr BSOUT						; ... dem Schläger löschen.
rts

; Ausgabe der Scores
scores_out:							; Punktezählung und -anzeige
	clc
	ldx #0							; Oben ...
	ldy #20							; ... mittig ...
	jsr PLOT						; ... Score
	lda punkte_player2				; ... Spieler 1 ...
	jsr BSOUT						; .... ausgeben.
	clc
	ldx #24							; Unten ...
	ldy #20							; ... mittig ...
	jsr PLOT						; ... Score ..
	lda punkte_player1				; ... Spieler 2 ...
	jsr BSOUT						; ... ausgeben.
rts	

; Löschen des Balls
ball_erase:							; Ball löschen
	clc
	ldx ball_y						; Ball von X- ...
	ldy ball_x						; ... und Y- ...
	jsr PLOT						; ... Position ...
	lda #32							; ... wieder ...
	jsr BSOUT						; ... löschen
rts

; Ausgeben des Balls
ball_draw:							; Ball zeichnen
	clc
	ldx ball_y						; Ball an X- ...
	ldy ball_x						; ... und Y-Position ...
	jsr PLOT						; ... platzieren ...
	lda #113						; ... und ...
	jsr BSOUT						; ... zeichnen.
rts 

play_sound:
	lda soundloop					; Ist noch ein Sound zu spielen?
	beq no_sound
	lda #15							; Lautestes ...
	sta SID+24						; ... Volume
	lda #0
	sta SID+5						; Attack
	lda #240
	sta SID+6						; Sustain
	ldy soundtune
	lda sound_table,y
	sta SID+1						; Frequenz (LB)
	lda sound_table,y+1
	sta SID							; Frequenz (HB)
	lda #%00100001					; Rechteck ...
	sta SID+4						; ... Wellenform
	jsr wait_sound
	no_sound:
rts	

wait_sound:						
	lda RND							; Timer in a laden
	hold:
		cmp RND						; Auf a+1 warten
	beq hold						; (dauert 1/60 Sekunde)
	dec soundloop
	lda soundloop					; Sound ...
	bne wait_sound_end				; ... fertig gespielt?
	sound_off:						; Dann alle SID-Register wieder auf 0 setzen
	ldy #$18
	wloop:
		lda #0
		sta SID,y
		dey
	bne wloop
	wait_sound_end:
rts

; TABELLEN
delta_table: 						; [ball_bounce]
	.byte 2,1,0,1,2					; Summanden und Minuenden für x-Ablenkung des Balls (-2,-1,0,1,2)
treffer_table: 						; [ball_bat]
	.byte 5,4,3,2,1					; Winkelzuweisung abhängig von der Schläger-Trefferzone
ball_table:							; [ball_wall]
	.byte 5,4,0,2,1					; Reflexionswinkel für Wand-Treffer: 1<->5, 2<->4, 0 existiert nicht
speed_table:						; [ball_speed]
	.byte 5,10,15,10,5				; vertikale Ballgeschiwndigkeit abhängig von Schläger-Trefferzone
sound_table:						; 16-Bit-Wert=Frequenz*2^24/Taktfrequenz (Taktfrequenz Europa: 17734472/18 = 0,985248 MHz)
	.byte $10,$5d					; Wall sound: 	Dauer 16 ms = 1/60 Sek., 	Frenquenz 246 Hz. = $105d 
	.byte $20,$a9					; Paddle sound: Dauer 16 ms = 1/60 Sek.,	Frenquenz 491 Hz. = $20a9
	.byte $20,$a9					; Score sound: 	Dauer 220 ms = 14/60 Sek.,	Frenquenz 491 Hz. = $20a9 

joywait:							; Experimentelle Warteschleife
	ldy #$ff
	joywait1:
		ldy #$ff
		joywait2:
			dey
			bne joywait2
		dex
		bne joywait1
rts